@@ -37,22 +37,48 @@ function(x,y,dpl=NULL,rw=NULL,eps=1e-9,sort=TRUE,
 # is referred to elsewhere as `the' rectangular window.
 if(exists("deldirMsgeDone",envir=EnvSupp)) suppressMsge <- TRUE
 if(!suppressMsge){
-    cat(paste("\n     PLEASE NOTE:  The components \"delsgs\" and \"summary\"",
-        "of the", "\n     object returned by deldir() are now",
-        "DATA FRAMES rather than","\n     matrices (as they were prior",
-        "to release 0.0-18).",    "\n     See help(\"deldir\").\n",
-        "\n     PLEASE NOTE: The process that deldir() uses for determining\n",
-        "    duplicated points has changed from that used in version\n",
-        "    0.0-9 of this package (and previously). See help(\"deldir\").\n\n"))
+    mess <- paste("\n     PLEASE NOTE:  The components \"delsgs\"",
+                  "and \"summary\" of the\n object returned by deldir()",
+                  "are now DATA FRAMES rather than\n matrices (as they",
+                  "were prior to release 0.0-18).\n See help(\"deldir\").\n",
+                  "\n     PLEASE NOTE: The process that deldir() uses for",
+                  "determining\n duplicated points has changed from that",
+                  "used in version\n 0.0-9 of this package (and previously).",
+                  "See help(\"deldir\").\n\n")
+    message(mess)
     assign("deldirMsgeDone","xxx",envir=EnvSupp)
 }
 
-# If the first argument is a list, extract components x and y (and
-# possibly z).
-if(is.list(x)) {
+# If the first argument is a data frame, extract the column
+# named "z", if there is one, to be the "z weights".  Remove
+# this column from the first argument.  Extract the column named
+# "x", if there is one, to be the "x" coordinates, otherwise take
+# the "x" coordinates to be the first column which is isn't named
+# "y".  Extract the column named "y", if there is one, to be the
+# "y" coordinates, otherwise take the "y" coordinates to be the
+# first column which is isn't named "x".
+if(is.data.frame(x)) {
+        if(ncol(x) < 2)
+            stop(paste("If \"x\" is a data frame it must have\n",
+                       "at least two columns.\n"))
+        j  <- match("z",names(x))
+        if(!is.na(j)) {
+            if(is.null(z)) z <- x[,j]
+            x <- x[,-j]
+        }
+        j <- match(c("x","y"),names(x))
+        if(all(is.na(j))) j <- 1:2
+        if(is.na(j[2])) j[2] <- if(j[1]==1) 2 else 1
+        if(is.na(j[1])) j[1] <- if(j[2]==1) 2 else 1
+        y <- x[,j[2]]
+        x <- x[,j[1]]
+
+# If the first argument is a list (but not a data frame) extract
+# components x and y (and possibly z).
+} else if(is.list(x)) {
 	if(all(!is.na(match(c('x','y'),names(x))))) {
 		y <- x$y
-                z <- if(!is.null(x$z)) x$z else z
+                z <- if(is.null(z)) x$z else z
 		x <- x$x
 	}
 	else {
@@ -71,8 +97,7 @@ if(haveZ) {
 }
 
 # If a data window is specified, get its corner coordinates
-# and truncate the data by this window.  Discard any constraint
-# segments either of whose endpoints fall outside this window.
+# and truncate the data by this window.
 if(!is.null(rw)) {
 	xmin <- rw[1]
 	xmax <- rw[2]
@@ -82,6 +107,7 @@ if(!is.null(rw)) {
 	if(length(drop)>0) {
 		x <- x[-drop]
 		y <- y[-drop]
+                if(haveZ) z <- z[-drop]
 		n <- length(x)
 	}
 }
@@ -107,24 +133,37 @@ if(!is.null(dpl)) {
 	dpts <- dumpts(x,y,dpl,rw)
 	x    <- dpts$x
 	y    <- dpts$y
+        ndm  <- length(x) - n
         if(haveZ) {
-		ndum <- length(x)-length(z)
 		if(!is.null(zdum)) {
-			if(length(zdum) != ndum)
+			if(length(zdum) != ndm)
 				stop("The z dummy points are of the wrong length.\n")
 		} else {
-			zdum <- rep(NA,ndum)
+			zdum <- rep(NA,ndm)
 		}
 		z <- c(z,zdum)
 	}
-}
+} else ndm <- 0
 
 # Eliminate duplicate points:
-iii <- !duplicatedxy(x,y)
-ndm <- sum(iii[-(1:n)])
-n   <- sum(iii[1:n])
-x   <- x[iii]
-y   <- y[iii]
+iii <- duplicatedxy(x,y)
+if(any(iii)) {
+    kkk <- !iii
+    ndm <- sum(kkk[-(1:n)])
+    n   <- sum(kkk[1:n])
+    if(haveZ) {
+        jjj <- duplicated(data.frame(x=x,y=y,z=z))
+        if(sum(jjj) < sum(iii)) {
+            whinge <- paste("There were different z \"weights\" corresponding to\n",
+                            "duplicated points.\n",sep="")
+            warning(whinge)
+        }
+        z   <- z[kkk]
+    }
+    x   <- x[kkk]
+    y   <- y[kkk]
+    ind.orig <- which(!iii)
+} else ind.orig <- seq_along(iii)
 
 # Make space for the total number of points (real and dummy) as
 # well as 4 ideal points and 4 extra corner points which get used
@@ -233,7 +272,8 @@ rw               <- round(rw,digits)
 
 # Aw' done!!!
 rslt <- list(delsgs=delsgs,dirsgs=dirsgs,summary=allsum,n.data=n,
-             n.dum=ndm,del.area=del.area,dir.area=dir.area,rw=rw)
+             n.dum=ndm,del.area=del.area,dir.area=dir.area,rw=rw,
+             ind.orig=ind.orig)
 class(rslt) <- 'deldir'
 if(plotit) {
 	plot(rslt,...)
